class DiscobaseError(Exception):
    """
    Base discobase exception class.
    """


class NotConnectedError(DiscobaseError):
    """
    The database is not connected.
    """


class DatabaseCorruptionError(DiscobaseError):
    """
    The database was corrupted somehow.
    """


class DatabaseStorageError(DiscobaseError):
    """
    Failed store something in the database.
    """


class DatabaseTableError(DiscobaseError):
    """
    Something is wrong with a `Table` type.
    """


class DatabaseLookupError(DiscobaseError):
    """
    Something went wrong with an entry lookup.
    """



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from typing import Any, Coroutine, TypeVar

import discord
from loguru import logger

__all__ = "gather_group", "GatherGroup", "free_fly"

T = TypeVar("T")


class GatherGroup:
    def __init__(self) -> None:
        self.tasks: list[asyncio.Task] = []

    def add(self, awaitable: Coroutine[Any, Any, T]) -> asyncio.Task[T]:
        async def inner_coro():
            while True:
                try:
                    return await awaitable
                except discord.HTTPException as e:
                    if e.code == 429:
                        logger.warning("Ratelimited! Retrying...")
                        await asyncio.sleep(0.1)

        task = asyncio.create_task(inner_coro())
        self.tasks.append(task)
        return task


# A partial reimplementation of asyncio.TaskGroup, but
# that's only on 3.11+ anyway.
@asynccontextmanager
async def gather_group():
    group = GatherGroup()

    try:
        yield group
    finally:
        logger.debug(f"Gathering tasks: {group.tasks}")
        await asyncio.gather(*group.tasks)


_TASKS = set()


def free_fly(coro: Coroutine[Any, Any, T]) -> asyncio.Task[T]:
    task = asyncio.create_task(coro)
    _TASKS.add(task)
    task.add_done_callback(_TASKS.discard)
    return task



from __future__ import annotations

import asyncio
from contextlib import asynccontextmanager
from typing import Coroutine, NoReturn, Type, TypeVar

import discord
from discord.ext import commands
from loguru import logger

from ._cursor import TableCursor
from .cogs.utility import Utility
from .cogs.visualization import Visualization
from .exceptions import (DatabaseCorruptionError, DatabaseTableError,
                         NotConnectedError)
from .table import Table

__all__ = ("Database",)

T = TypeVar("T", bound=Type[Table])


class Database:
    """
    Top level class representing a Discord
    database bot controller.
    """

    def __init__(
        self,
        name: str,
        logging: bool = False,
    ) -> None:
        """
        Args:
            name: Name of the Discord server that will be used as the database.
            logging: Whether to enable logging.
        """
        if logging:
            logger.enable("discobase")
        else:
            logger.disable("discobase")

        self.name = name
        """Name of the Discord-database server."""
        self.bot = commands.Bot(
            intents=discord.Intents.all(),
            command_prefix="!",
        )
        """discord.py `Bot` instance."""
        self.guild: discord.Guild | None = None
        """discord.py `Guild` used as the database server."""
        self.tables: dict[str, type[Table]] = {}
        """Dictionary of `Table` objects attached to this database."""
        self.open: bool = False
        """Whether the database is connected."""
        self._metadata_channel: discord.TextChannel | None = None
        """discord.py `TextChannel` that acts as the metadata channel."""
        self._database_cursors: dict[str, TableCursor] = {}
        """A dictionary containing all of the table `Metadata` entries"""
        self._task: asyncio.Task[None] | None = None
        self.bot.db = self  # type: ignore
        # We need to keep a strong reference to the free-flying
        # task
        self._setup_event = asyncio.Event()
        self._internal_setup_event = asyncio.Event()
        self._on_ready_exc: BaseException | None = None

        # Here be dragons: https://github.com/ZeroIntensity/discobase/issues/49
        #
        # `on_ready` in discord.py swallows all exceptions, which
        # goes against some connect-and-disconnect behavior
        # that we want to allow in discobase.
        #
        # We need to store the exception, and then raise in wait_ready()
        # in order to properly handle it, otherwise the discord.py
        # logger just swallows it and pretends nothing happened.
        #
        # This also caused a deadlock with _setup_event, which caused
        # CI to run indefinitely.
        @self.bot.event
        @logger.catch(reraise=True)
        async def on_ready() -> None:
            try:
                await self.bot.add_cog(Utility(self.bot))
                await self.bot.add_cog(Visualization(self.bot))
                await self.init()
            except BaseException as e:
                await self.bot.close()
                if self._task:
                    self._task.cancel("bot startup failed")

                self._setup_event.set()
                self._on_ready_exc = e
                raise  # This is swallowed!

    def _not_connected(self) -> NoReturn:
        """
        Complain about the database not being connected.

        Generally speaking, this is called when `guild` or something
        other is `None`.
        """

        raise NotConnectedError(
            "not connected to the database, did you forget to login?"
        )

    async def _metadata_init(self) -> discord.TextChannel:
        """
        Find the metadata channel.
        If it doesn't exist, this method creates one.

        Returns:
            discord.TextChannel: The metadata channel, either created or found.
        """
        metadata_channel_name = "_dbmetadata"
        found_channel: discord.TextChannel | None = None
        assert self.guild is not None

        for channel in self.guild.text_channels:
            if channel.name == metadata_channel_name:
                found_channel = channel
                logger.info("Found metadata channel!")
                break

        return found_channel or await self.guild.create_text_channel(
            name=metadata_channel_name
        )

    # This needs to be async for use in gather()
    async def _set_open(self) -> None:
        logger.debug("_set_open waiting on internal setup event")
        await self._internal_setup_event.wait()
        logger.debug(
            "Internal setup event dispatched! Database has been marked as open."  # noqa
        )
        self.open = True
        # See https://github.com/ZeroIntensity/discobase/issues/68
        #
        # If `wait_ready()` is never called, then the error does not propagate.
        if self._on_ready_exc:
            raise self._on_ready_exc

    async def init(self) -> None:
        """
        Initializes the database server.

        Generally, you don't want to call this manually, but
        this is considered to be a public interface.
        """
        logger.info("Waiting until bot is logged in.")
        await self.bot.wait_until_ready()
        logger.info("Bot is ready!")
        found_guild: discord.Guild | None = None
        for guild in self.bot.guilds:
            if guild.name == self.name:
                found_guild = guild
                break

        if not found_guild:
            logger.info("No guild found, making one.")
            self.guild = await self.bot.create_guild(name=self.name)
        else:
            logger.info("Found an existing guild.")
            self.guild = found_guild

        # Unlock database, but don't wakeup the user.
        self._internal_setup_event.set()
        await self.build_tables()
        # At this point, the database is marked as "ready" to the user.
        self._setup_event.set()

        assert self._metadata_channel is not None
        logger.info(
            f"Invite to server: {await self._metadata_channel.create_invite()}"
        )
        logger.info("Syncing slash commands, this might take a minute.")
        logger.debug(f"Synced slash commands: {await self.bot.tree.sync()}")

    async def build_tables(self) -> None:
        """
        Generate all internal metadata and construct tables.

        Calling this manually is useful if e.g. you want
        to load tables *after* calling `login` (or `login_task`, or
        `conn`, same difference.)

        This method is safe to call multiple times.

        Example:
            ```py
            import asyncio
            import discobase

            async def main():
                db = discobase.Database("My database")
                db.login_task("My bot token")

                @db.table
                class MyLateTable(discobase.Table):
                    something: str

                # Load MyLateTable into database
                await db.build_tables()

            asyncio.run(main())
            ```
        """
        if not self.guild:
            self._not_connected()

        self._metadata_channel = await self._metadata_init()
        tasks = [
            asyncio.ensure_future(
                TableCursor.create_table(
                    table,
                    self._metadata_channel,
                    self.guild,
                )
            )
            for table in self.tables.values()
        ]
        logger.debug(f"Creating tables with gather(): {tasks}")
        await asyncio.gather(*tasks)

    async def wait_ready(self) -> None:
        """
        Wait until the database is ready.
        """
        logger.info("Waiting until the database is ready.")
        await self._setup_event.wait()
        logger.info("Done waiting!")
        # See #49, we need to propagate errors in `on_ready` here.
        if self._on_ready_exc:
            logger.error("on_ready() failed, propagating now.")
            raise self._on_ready_exc

    def _find_channel(self, cid: int) -> discord.TextChannel:
        # TODO: Implement caching for this function.
        if not self.guild:
            self._not_connected()

        index_channel = [
            channel for channel in self.guild.channels if channel.id == cid
        ][0]

        if not isinstance(index_channel, discord.TextChannel):
            raise DatabaseCorruptionError(
                f"expected {index_channel!r} to be a TextChannel"
            )

        logger.debug(f"Found channel ID {cid}: {index_channel!r}")
        return index_channel

    async def clean(self) -> None:
        """
        Perform a full clean of the database.

        This method erases all metadata, all entries, and all tables. After
        calling this, a server loses any trace of the database ever being
        there.

        Note that this does *not* clean the existing tables from memory, but
        instead just marks them all as not ready.

        This action is irreversible.
        """
        if not self._metadata_channel:
            self._not_connected()

        logger.info("Cleaning the database!")

        coros: list[Coroutine] = []
        for table, cursor in self._database_cursors.items():
            metadata = cursor.metadata
            logger.info(f"Cleaning table {table}")
            table_channel = self._find_channel(metadata.table_channel)
            coros.append(table_channel.delete())

            for cid in metadata.index_channels.values():
                channel = self._find_channel(cid)
                coros.append(channel.delete())

        for schema in self.tables.values():
            schema.__disco_cursor__ = None

        logger.debug(f"Gathering deletion coros: {coros}")
        await asyncio.gather(*coros)
        logger.info("Deleting database metadata.")
        self._database_cursors = {}
        await self._metadata_channel.delete()

    async def login(self, bot_token: str) -> None:
        """
        Start running the bot.

        Args:
            bot_token: Discord API bot token to log in with.
        """
        if self.open:
            raise RuntimeError(
                "connection is already open, did you call login() twice?"
            )

        # We use _set_open() with a gather to keep a finer link
        # between the `open` attribute and whether it's actually
        # running.
        await asyncio.gather(self.bot.start(token=bot_token), self._set_open())

    def login_task(self, bot_token: str) -> asyncio.Task[None]:
        """
        Call `login()` as a free-flying task, instead of
        blocking the event loop.

        Note that this method stores a reference to the created
        task object, allowing it to be truly "free-flying."

        Args:
            bot_token: Discord API bot token to log in to.

        Returns:
            asyncio.Task[None]: The created `asyncio.Task` object.
                Note that the database will store this internally, so you
                don't have to worry about losing the reference. By default,
                this task will never get `await`ed, so this function will not
                keep the event loop running. If you want to keep the event loop
                running, make sure to `await` the returned task object later.

        Example:
            ```py
            import asyncio
            import os

            import discobase


            async def main():
                db = discobase.Database("test")
                task = await db.login_task("...")
                await db.wait_ready()
                # ...
                await task  # Keep the event loop running

            asyncio.run(main())
            ```
        """
        task = asyncio.create_task(self.login(bot_token))
        self._task = task
        return task

    async def close(self) -> None:
        """
        Close the bot connection.
        """
        if not self.open:
            # If something went wrong in startup, for example, then
            # we need to release the setup lock.
            self._setup_event.set()
            raise ValueError(
                "cannot close a connection that is not open",
            )
        self.open = False
        await self.bot.close()

    @asynccontextmanager
    async def conn(self, bot_token: str):
        """
        Connect to the bot under a context manager.
        This is the recommended method to use for logging in.

        Args:
            bot_token: Discord API bot token to log in to.

        Returns:
            AsyncGeneratorContextManager: An asynchronous context manager.
                See `contextlib.asynccontextmanager` for details.

        Example:
            ```py
            import asyncio
            import os

            import discobase


            async def main():
                db = discobase.Database("test")
                async with db.conn(os.getenv("BOT_TOKEN")):
                    ...  # Your database code


            asyncio.run(main())
            ```
        """
        try:
            self.login_task(bot_token)
            await self.wait_ready()
            yield
        finally:
            if self.open:  # Something could have gone wrong
                await self.close()

    def table(self, clas: T) -> T:
        """
        Mark a `Table` type as part of a database.
        This method is meant to be used as a decorator.

        Args:
            clas: Type object to attach.

        Example:
            ```py
            import discobase

            db = discobase.Database("My database")

            @db.table
            class MySchema(discobase.Table):
                foo: int
                bar: str

            # ...
            ```

        Returns:
            Type[Table]: The same object passed to `clas` -- this is in order
                to allow use as a decorator.
        """
        if not issubclass(clas, Table):
            raise DatabaseTableError(
                f"{clas} is not a subclass of discobase.Table, did you forget it?",  # noqa
            )

        clas.__disco_name__ = clas.__name__.lower()
        if clas.__disco_name__ in self.tables:
            raise DatabaseTableError(f"table {clas.__name__} already exists")

        if clas.__disco_database__ is not None:
            raise DatabaseTableError(
                f"{clas!r} can only be attached to one database"
            )

        clas.__disco_database__ = self

        # This is up for criticism -- instead of using Pydantic's
        # `model_fields` attribute, we invent our own `__disco_keys__` instead.
        #
        # Partially, this is due to the fact that we want `__disco_keys__` to
        # be, more or less, stable throughout the codebase.
        #
        # However, I don't think Pydantic would mess with `model_fields`, as
        # that's a public API, and hence why this could possibly be
        # considered as bad design.
        for field in clas.model_fields:
            clas.__disco_keys__.add(field)

        self.tables[clas.__disco_name__] = clas
        return clas



from __future__ import annotations

import asyncio
import hashlib
from base64 import urlsafe_b64decode, urlsafe_b64encode
from collections.abc import Iterable
from datetime import timedelta
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, List, Optional

import discord
from aiocache import cached
from discord.utils import snowflake_time, time_snowflake
from loguru import logger
from pydantic import BaseModel, ValidationError

from ._metadata import Metadata
from .exceptions import (DatabaseCorruptionError, DatabaseLookupError,
                         DatabaseStorageError)

if TYPE_CHECKING:
    from .table import Table

__all__ = ("TableCursor",)


class _Record(BaseModel):
    content: str
    """Base64 encoded Pydantic model dump of the record."""

    @classmethod
    def from_data(cls, data: Table) -> _Record:
        logger.debug(f"Generating a _Record from data: {data}")
        return _Record(
            content=urlsafe_b64encode(  # Record JSON data is stored in base64
                data.model_dump_json().encode("utf-8"),
            ).decode("utf-8"),
        )

    def decode_content(self, record: Table | type[Table]) -> Table:
        return record.model_validate_json(urlsafe_b64decode(self.content))


class _IndexableRecord(BaseModel):
    key: int
    """Hashed value of the key."""
    record_ids: List[int]
    """Message IDs of the records that correspond to this key."""
    next_value: Optional[_IndexableRecord] = None
    """
    Temporary placeholder value for the next entry.
    Only for use in resizing.
    """

    @classmethod
    def from_message(cls, message: str) -> _IndexableRecord | None:
        """
        Generate an `_IndexableRecord` instance from message content.

        Args:
            message: Message content to parse as JSON.

        Returns:
            _IndexableRecord | None: An `_IndexableRecord` instance, or `None`,
                if the message was `null`.
        """
        logger.debug(f"Parsing {message} into an _IndexableRecord")
        try:
            return (
                cls.model_validate_json(message) if message != "null" else None
            )
        except ValidationError as e:
            raise DatabaseCorruptionError(
                f"got bad _IndexableRecord entry: {message}"
            ) from e


class _HashTransport:
    """
    Hacky object to use `hash()` for tuples and dictionaries
    that retains the value between interpreters.
    """

    def __init__(self, hash_num: int) -> None:
        self.hash_num = hash_num

    def __hash__(self) -> int:
        return self.hash_num


class TableCursor:
    def __init__(
        self,
        metadata: Metadata,
        metadata_channel: discord.TextChannel,
        guild: discord.Guild,
    ) -> None:
        self.metadata = metadata
        self.metadata_channel = metadata_channel
        self.guild = guild

    @lru_cache
    def _find_channel(self, channel_id: int) -> discord.TextChannel:
        for channel in self.guild.channels:
            if channel.id != channel_id:
                continue

            if not isinstance(channel, discord.TextChannel):
                raise DatabaseCorruptionError(
                    f"{channel!r} is not a TextChannel"
                )

            return channel

        raise DatabaseCorruptionError(
            f"could not find channel with id {channel_id}"
        )

    async def _find_collision_message(
        self,
        channel: discord.TextChannel,
        index: int,
        *,
        search_func: Callable[[str], bool] = lambda s: s == "null",
    ) -> discord.Message:
        """
        Search for a message via a worst-case O(n) search in the event
        of a hash collision.

        Args:
            channel: Index channel to search.
            index: The index to start at.
            search_func: Function to check if the message content is good.

        Returns:
            discord.Message: The message that satisfies search_func
        """
        logger.debug(
            f"Looking up hash collision entry using search function: {search_func}"  # noqa
        )
        offset: int = index
        while True:
            if (offset + 1) >= self.metadata.max_records:
                logger.debug("We need to wrap around the table.")
                offset = 0
            else:
                offset += 1

            if offset == index:
                raise DatabaseCorruptionError(
                    f"index channel {channel!r} has no free messages, table was likely not resized."  # noqa
                )

            message = await self._lookup_message(
                channel,
                offset,
            )
            logger.debug(
                f"Hash collision search at index: {offset} {message=}",
            )
            if search_func(message.content):
                logger.debug(
                    f"Done searching for collision message: {message.content}"
                )
                return message

    async def _edit_message(
        self,
        channel: discord.TextChannel,
        mid: int,
        content: str,
    ) -> None:
        """
        Edit a message given the channel, message ID, and content.

        This should *not* be used over `discord.Message.edit`, it's simply
        a handy utility to use when you only have the message ID.
        """
        editable_message = await channel.fetch_message(mid)
        logger.debug(f"Editing message (ID {mid}) to {content}")
        await editable_message.edit(content=content)

    def _to_index(self, value: int) -> int:
        """
        Generate an index from a hash number based
        on the metadata's `max_records`.

        Args:
            value: Integer hash, positive or negative.

        Returns:
            int: Index in range of `metadata.max_records`.
        """
        index = (value & 0x7FFFFFFF) % self.metadata.max_records
        logger.debug(
            f"Hashed value {value} turned into index: {index} ({self.metadata.max_records=})"  # noqa
        )
        return index

    @lru_cache()
    def _hash(
        self,
        value: Any,
    ) -> int:
        """
        Hash the field into an integer.

        Args:
            value: Any discobase-hashable object.

        Returns:
            int: An integer, positive or negative, representing a unique hash.
                This is always the same thing across programs.
        """
        logger.debug(f"Hashing object: {value!r}")
        if isinstance(value, str):
            hashed_str = int(
                hashlib.sha1(value.encode("utf-8")).hexdigest(),
                16,
            )
            logger.debug(f"Hashed string {value!r} into {hashed_str}")
            return hashed_str
        elif isinstance(value, dict):
            transport: dict[_HashTransport, _HashTransport] = {}

            for k, v in value.items():
                transport[_HashTransport(self._hash(k))] = _HashTransport(
                    self._hash(v)
                )

            hashed_dict = hash(transport)
            logger.debug(f"Hashed dictionary {value!r} into {hashed_dict}")
            return hashed_dict
        elif isinstance(value, Iterable):
            hashes: list[_HashTransport] = []
            for item in value:
                hashes.append(_HashTransport(self._hash(item)))

            hashed_tuple = hash(tuple(hashes))
            logger.debug(f"Hashed iterable {value!r} into {hashed_tuple}")
            return hashed_tuple
        elif isinstance(value, int):
            return value
        else:
            raise DatabaseStorageError(f"unhashable: {value!r}")

    def _as_hashed(
        self,
        value: Any,
    ) -> tuple[int, int]:
        """
        Get the hash number and index for `value`.
        """
        hashed = self._hash(value)
        return hashed, self._to_index(hashed)

    @cached()
    async def _lookup_message_impl(
        self,
        channel: discord.TextChannel,
        index: int,
    ) -> discord.Message:
        """
        The *implementation* of looking up a message by
        it's index in the table. You need to call `fetch()`
        on the result of this function due to caching.

        Args:
            channel: Index channel to search.
            index: Index to get.

        Returns:
            discord.Message: The found message.

        Raises:
            DatabaseCorruptionError: Could not find the index.
        """
        metadata = self.metadata
        logger.debug(f"Looking up message: {index}")
        for timestamp, rng in metadata.time_table.items():
            start: int = rng[0]
            end: int = rng[1]
            if index not in range(
                start, end
            ):  # Pydantic doesn't support ranges
                continue

            logger.debug(f"In range: {start} - {end}")
            current_index: int = 0
            async for msg in channel.history(
                limit=end - start,
                before=snowflake_time(timestamp),
            ):
                if current_index == (index - start):
                    logger.debug(f"{msg} found at index {current_index}")
                    return msg
                current_index += 1

            raise DatabaseCorruptionError(
                f"range for {timestamp} in table {metadata.name} does not contain index {index}"  # noqa
            )

        raise DatabaseCorruptionError(
            f"message index out of range for table {metadata.name}: {index}"
        )

    async def _lookup_message(
        self,
        channel: discord.TextChannel,
        index: int,
    ) -> discord.Message:
        """
        Lookup a message by it's index in the table.

        Args:
            channel: Index channel to search.
            index: Index to get.

        Returns:
            discord.Message: The found message.

        Raises:
            DatabaseCorruptionError: Could not find the index.
        """
        # We need to refetch it for the latest content.
        return await (await self._lookup_message_impl(channel, index)).fetch()

    async def _resize_hash(
        self,
        index_channel: discord.TextChannel,
        amount: int,
    ) -> int:
        """
        Increases the hash for `index_channel` by amount

        Args:
            index_channel: the channel that contains index data for a database
            amount: the amount to increase the size by

        Returns:
            int: snowflake representation of when the last message of the
                resize was created
        """
        last_message: discord.Message | None = None

        # Here be dragons: ratelimit makes gathering this actually worse.
        for _ in range(amount):
            last_message = await index_channel.send("null", silent=True)

        if not last_message:
            raise DatabaseCorruptionError("last_message is None somehow")
        # 5 seconds, per the Discord ratelimit
        last_timestamp = timedelta(seconds=5) + last_message.created_at
        return time_snowflake(last_timestamp)

    async def _resize_channel(
        self,
        channel: discord.TextChannel,
    ) -> None:
        """
        The implementation of resizing a channel. This method assumes
        that `self.metadata.max_records` has already been doubled.

        This is meant for use in `gather()`, for optimal performance.

        Args:
            channel: Index channel to resize.
        """
        metadata = self.metadata
        logger.debug(
            f"Resizing channel: {channel!r} for table {metadata.name}",
        )
        old_size: int = metadata.max_records // 2
        timestamp_snowflake = await self._resize_hash(channel, old_size)
        rng = (
            old_size,
            metadata.max_records,
        )

        for snowflake, time_range in metadata.time_table.copy().items():
            # We only want one time stamp for the range, this forces
            # the latest one to always be used -- that's a good thing,
            # we don't want to risk having messages from the previous range
            # in this one.
            if time_range == rng:
                del metadata.time_table[snowflake]

        metadata.time_table[timestamp_snowflake] = rng
        # Now, we have to move everything into the correct position.
        #
        # Note that this shouldn't put everything into memory, as
        # each previous iteration will be freed -- this is good
        # for scalability.
        #
        # Due to Discord's ratelimit, gathering the coros in this loop
        # is actually a bad idea.
        async for msg in channel.history(
            limit=old_size,
            oldest_first=True,
        ):
            # msg = await channel.fetch_message(msg.id)
            record = _IndexableRecord.from_message(msg.content)
            if not record:
                continue

            new_index: int = self._to_index(record.key)
            target = await self._lookup_message(
                channel,
                new_index,
            )

            next_record = _IndexableRecord.from_message(target.content)
            inplace: bool = True
            overwrite: bool = True

            if next_record:
                if next_record.next_value:
                    logger.info("Hash collision in resize!")
                    target = await self._find_collision_message(
                        channel,
                        new_index,
                    )
                    # `inplace` is True, so we fall
                    # through to the inplace edit.
                    #
                    # To be fair, I'm not too sure if this is
                    # the best approach, this might be worth
                    # refactoring in the future.
                else:
                    logger.info("Updating record at the new index.")
                    inplace = False
                    logger.debug(
                        f"{next_record} marked as the next value location ({target.id=})"  # noqa
                    )

                    if record.next_value:
                        record.next_value = None
                        # Here be dragons: if we overwrite the `next_value`
                        # with `None` to prevent a doubly-nested copy in
                        # the JSON, we have to mark this message to *not*
                        # be overwritten, otherwise we lose that data.
                        overwrite = False

                    next_record.next_value = record
                    content = next_record.model_dump_json()
                    logger.debug(f"Editing {target.content} to {content}")
                    await target.edit(content=content)

            if inplace:
                # In case of a hash collision, we want to mark
                # this as having a `next_value`, so it doesn't get
                # overwritten.
                #
                # We copy this to prevent a recursive model dump.
                if record.next_value:
                    record.next_value = None
                    overwrite = False

                copy = record.model_copy()
                copy.next_value = record
                logger.info(
                    "Target index does not have an entry, updating in-place."  # noqa
                )
                content = copy.model_dump_json()
                logger.debug(f"Editing in-place null to {content}")
                assert target.content == "null"
                await target.edit(content=content)

            # Technically speaking, the index could
            # remain the same. We need to check for that.
            if (not record.next_value) and (target != msg) and overwrite:
                await msg.edit(content="null")

        # Finally, all the next_value attributes have been set, we can
        # go through and update each record.
        #
        # The overall algorithm is O(2n), but it's much more scalable
        # than trying to put the entire table into memory in order to
        # resize it.
        #
        # This algorithm is pretty much infinitely scalable
        # in terms of memory, but we're limited by Discord's ratelimit.
        async for msg in channel.history(
            limit=metadata.max_records,
            oldest_first=True,
        ):
            record = _IndexableRecord.from_message(msg.content)
            if not record:
                continue

            logger.debug(f"Handling movement of {record!r}")
            if not record.next_value:
                raise DatabaseCorruptionError(
                    "all existing records after resize should have next_value",  # noqa
                )

            if record.next_value.next_value:
                raise DatabaseCorruptionError(
                    f"doubly nested next_value found: {record.next_value.next_value!r} in {record!r}"  # noqa
                )

            content = record.next_value.model_dump_json()
            logger.debug(f"Replacing {msg.content} with {content}")
            await msg.edit(content=content)

    async def _resize_table(self) -> None:
        """
        Resize all the index channels in a table.
        """
        metadata = self.metadata
        metadata.max_records *= 2
        logger.info(
            f"Resizing table {metadata.name} to {metadata.max_records}"  # noqa
        )
        await asyncio.gather(
            *[
                self._resize_channel(self._find_channel(cid))
                for cid in metadata.index_channels.values()
            ]
        )

        # Dump the new metadata
        await self._edit_message(
            self.metadata_channel,
            metadata.message_id,
            metadata.model_dump_json(),
        )
        logger.info(
            f"Table {metadata.name} is now of size {metadata.max_records}"
        )

    async def _inc_records(self) -> None:
        """
        Increment the `current_records` number on the
        target metadata. This resizes the table if the maximum
        size is reached.
        """
        metadata = self.metadata
        metadata.current_records += 1
        if metadata.current_records > metadata.max_records:
            logger.info("The table is full! We need to resize it.")
            await self._resize_table()

        await self._edit_message(
            self.metadata_channel,
            metadata.message_id,
            metadata.model_dump_json(),
        )

    async def _write_index_record(
        self,
        channel: discord.TextChannel,
        index: int,
        hashed: int,
        record_id: int,
    ) -> None:
        """
        Write an index record to the specified channel, using
        a known hash and index.

        Args:
            channel: Target index channel to store the index record at.
            index: Index to store the record at in the table.
            hashed: Integer hash of the original value e.g. from `_hash`.
            record_id: Message ID of the record in the main table.
        """
        entry_message: discord.Message = await self._lookup_message(
            channel,
            index,
        )
        serialized_content = _IndexableRecord.from_message(
            entry_message.content
        )

        if not serialized_content:
            logger.info("This is a null entry, we can just update in place.")
            await self._inc_records()
            message_content = _IndexableRecord(
                key=hashed,
                record_ids=[
                    record_id,
                ],
            )
            await entry_message.edit(content=message_content.model_dump_json())
        elif serialized_content.key == hashed:
            # See https://github.com/ZeroIntensity/discobase/issues/50
            #
            # We don't want to call _inc_records() here, because we aren't
            # using up a `null` space.
            logger.info("This already exists, let's append to the data.")
            serialized_content.record_ids.append(record_id)
            await entry_message.edit(
                content=serialized_content.model_dump_json()
            )
        else:
            logger.info("Hash collision!")
            await self._inc_records()
            index_message = await self._find_collision_message(
                channel,
                index,
            )
            collision_entry = _IndexableRecord(
                key=hashed,
                record_ids=[
                    record_id,
                ],
            )
            await index_message.edit(content=collision_entry.model_dump_json())

    async def add_record(self, record: Table) -> discord.Message:
        """
        Writes a record to an existing table.

        Args:
            record: The record object being written to the table

        Returns:
            discord.Message: The `discord.Message` that contains the new entry.
        """

        metadata = self.metadata
        record_data = _Record.from_data(record)
        main_table: discord.TextChannel = self._find_channel(
            metadata.table_channel
        )
        message = await main_table.send(
            record_data.model_dump_json(), silent=True
        )

        for field, value in record.model_dump().items():
            channel = self._find_channel(
                metadata.index_channels[f"{record.__disco_name__}_{field}"]
            )
            hashed_field, target_index = self._as_hashed(value)
            await self._write_index_record(
                channel,
                target_index,
                hashed_field,
                message.id,
            )

        return await message.edit(content=record_data.model_dump_json())

    async def update_record(self, record: Table) -> discord.Message:
        """
        Updates an existing record in a table.

        Args:
            record: The record object being written to the table

        Returns:
            discord.Message: The `discord.Message` that contains the new entry.
        """
        if record.__disco_id__ == -1:
            # Sanity check
            raise DatabaseCorruptionError("record must have an id to update")

        metadata = self.metadata
        main_table: discord.TextChannel = self._find_channel(
            metadata.table_channel
        )
        msg = await main_table.fetch_message(record.__disco_id__)
        current = _Record.model_validate_json(msg.content).decode_content(
            record
        )
        await msg.edit(content=_Record.from_data(record).model_dump_json())

        for new, old in zip(
            record.model_dump().items(),
            current.model_dump().items(),
        ):
            field = new[0]
            if field != old[0]:
                raise DatabaseCorruptionError(
                    f"field name {field} does not match {old[0]}"
                )

            new_value = new[1]
            old_value = old[1]
            if new_value == old_value:
                logger.info("Nothing changed.")
                continue

            channel = self._find_channel(
                metadata.index_channels[f"{record.__disco_name__}_{field}"]
            )
            hashed_field, target_index = self._as_hashed(new_value)
            await self._write_index_record(
                channel,
                target_index,
                hashed_field,
                msg.id,
            )

            old_index = self._to_index(self._hash(old_value))
            old_msg = await self._lookup_message(channel, old_index)
            old_record = _IndexableRecord.from_message(old_msg.content)
            if not old_record:
                raise DatabaseCorruptionError(
                    "got null record somehow",
                )

            if len(old_record.record_ids) == 1:
                logger.info("We can nullify this entry.")
                await old_msg.edit(content="null")
                self.metadata.current_records -= 1
            else:
                logger.info(
                    "There are other entries with this value, only remove this ID."  # noqa
                )
                old_record.record_ids.remove(msg.id)
                await old_msg.edit(content=old_record.model_dump_json())

        return msg

    async def find_records(
        self,
        table: type[Table],
        query: dict[str, Any],
    ) -> list[Table]:
        """
        Find a record based on the specified field values.

        Args:
            table: Table type to find records for.
            query: Dictionary containing field-value pairs.

        Returns:
            list[Table]: A list of `Table` objects (or really, a list of
                objects that inherit from `Table`), with the appropriate values
                specified by `query`.
        """
        metadata = self.metadata
        name = table.__disco_name__
        sets_list: list[set[int]] = []

        logger.debug(f"Looking for query {query!r} in {name}")
        for field, value in query.items():
            if field not in metadata.keys:
                raise DatabaseLookupError(
                    f"table {metadata.name} has no field {field}"
                )

            channel = self._find_channel(
                metadata.index_channels[f"{name}_{field}"]
            )

            hashed_field, target_index = self._as_hashed(value)
            entry_message = await self._lookup_message(
                channel,
                target_index,
            )

            serialized_content = _IndexableRecord.from_message(
                entry_message.content
            )

            if not serialized_content:
                logger.info("Nothing was found.")
                continue

            if serialized_content.key == hashed_field:
                logger.debug(f"Key matches hash! {serialized_content}")
                sets_list.append(set(serialized_content.record_ids))
            else:
                # Hash collision!
                def find_hash(message: str | None) -> bool:
                    if not message:
                        return False

                    index_record = _IndexableRecord.from_message(message)
                    if not index_record:
                        return False

                    return index_record.key == hashed_field

                entry = await self._find_collision_message(
                    channel,
                    target_index,
                    search_func=find_hash,
                )

                rec = _IndexableRecord.from_message(entry.content)
                logger.debug(f"Found hash collision index entry: {rec}")  # noqa
                if not rec:
                    # This shouldn't be possible, considering the
                    # search function explicitly disallows that.
                    raise DatabaseCorruptionError(
                        "search function found null entry somehow"
                    )

                sets_list.append(set(rec.record_ids))

        if not query:
            logger.info("Query is empty, finding all entries!")
            channel = self._find_channel(metadata.table_channel)
            async for msg in channel.history(limit=None):
                logger.debug(f"Found message in channel: {msg}")
                sets_list.append({msg.id})

        main_table = self._find_channel(metadata.table_channel)
        if not isinstance(main_table, discord.TextChannel):
            raise DatabaseCorruptionError(
                f"expected {main_table!r} to be a TextChannel"
            )

        logger.debug(f"Got IDs: {sets_list}")
        records: list[Table] = []

        for record_ids in sets_list:
            for record_id in record_ids:
                message = await main_table.fetch_message(record_id)
                record = _Record.model_validate_json(message.content)
                entry = record.decode_content(table)
                entry.__disco_id__ = message.id
                records.append(entry)

        return records

    async def _gen_key_channel(
        self,
        table: str,
        key_name: str,
        *,
        initial_size: int = 4,
    ) -> tuple[str, int, int]:
        """
        Generate a key channel from the given information.
        This does not check if it exists.

        Args:
            table: Processed channel name of the table.
            key_name: Name of the key, per `__disco_keys__`.
            initial_size: Equivalent to `initial_size` in `create_table`.

        Returns:
            tuple[int, int, int]: Tuple containing the channel name,
                the ID of the created channel, and the snowflake time of the
                last message created in the hash table.
        """
        # Key channels are stored in
        # the format of <table_name>_<field_name>
        index_channel = await self.guild.create_text_channel(
            f"{table}_{key_name}"
        )
        logger.debug(f"Generated key channel: {index_channel}")
        last_message_snowflake = await self._resize_hash(
            index_channel, initial_size
        )
        return index_channel.name, index_channel.id, last_message_snowflake

    @classmethod
    async def create_table(
        cls,
        table: type[Table],
        metadata_channel: discord.TextChannel,
        guild: discord.Guild,
        initial_size: int = 4,
    ) -> TableCursor:
        """
        Creates a new table and all index tables that go with it.
        This writes the table metadata.

        If the table already exists, this method does (almost) nothing.

        Args:
            table: Table schema to create channels for.
            initial_hash_size: the size the index hash tables should start at.

        Returns:
            TableCursor: An object used to manage a table
        """

        logger.debug(f"create_table called with table: {table!r}")
        name = table.__disco_name__
        existing_metadata: Metadata | None = None

        async for msg in metadata_channel.history(limit=None):
            try:
                parsed_meta = Metadata.model_validate_json(msg.content)
            except ValidationError as e:
                raise DatabaseCorruptionError("got invalid metadata") from e

            if parsed_meta.name == name:
                logger.debug(
                    f"Found existing metadata for table {name}: {parsed_meta}"
                )
                existing_metadata = parsed_meta
                break

        if existing_metadata and (
            set(existing_metadata.keys) != table.__disco_keys__
        ):
            logger.error(
                f"stored keys: {', '.join(existing_metadata.keys)} -- table keys: {', '.join(table.__disco_keys__)}"  # noqa
            )
            raise DatabaseCorruptionError(f"schema for table {name} changed")

        matching: list[str] = []
        for channel in guild.channels:
            for key in table.__disco_keys__:
                if channel.name == f"{name}_{key}":
                    matching.append(key)

        if existing_metadata and matching:
            if not len(matching) == len(table.__disco_keys__):
                raise DatabaseCorruptionError(
                    f"only some key channels exist: {', '.join(matching)}",
                )

            logger.info(f"Table is already set up: {table.__disco_name__}")
            cursor = TableCursor(existing_metadata, metadata_channel, guild)
            table.__disco_cursor__ = cursor
            return cursor

        logger.info(f"Building table: {table.__disco_name__}")

        # The primary table holds the actual records
        primary_table = await guild.create_text_channel(name)
        logger.debug(f"Generated primary table: {primary_table!r}")

        metadata = Metadata(
            name=name,
            keys=tuple(table.__disco_keys__),
            table_channel=primary_table.id,
            index_channels={},
            current_records=0,
            max_records=initial_size,
            time_table={},
            message_id=0,
        )
        self = TableCursor(metadata, metadata_channel, guild)
        timestamp_snowflake: int | None = None

        index_channels: dict[str, int] = {}
        # This is ugly, but this is fast: we generate
        # the key channels in parallel.
        for data in await asyncio.gather(
            *[
                self._gen_key_channel(
                    name,
                    key_name,
                    initial_size=initial_size,
                )
                for key_name in table.__disco_keys__
            ]
        ):
            channel_name, channel_id, timestamp_snowflake = data
            index_channels[channel_name] = channel_id

        assert timestamp_snowflake is not None
        metadata.time_table = {timestamp_snowflake: (0, initial_size)}
        metadata.index_channels = index_channels
        message = await self.metadata_channel.send(
            metadata.model_dump_json(), silent=True
        )

        table.__disco_cursor__ = self
        # Since Discord generates the message ID, we have to do these
        # message editing shenanigans.
        metadata.message_id = message.id
        await message.edit(content=metadata.model_dump_json())
        logger.debug(f"Generated table metadata: {metadata!r}")
        return self

    async def delete_record(self, record: Table) -> None:
        """
        Deletes an existing record in a table.

        Args:
            record: The record object being deleted from the table.
        """
        if record.__disco_id__ == -1:
            # Sanity check
            raise DatabaseCorruptionError("record must have an id to update")

        metadata = self.metadata
        main_table: discord.TextChannel = self._find_channel(
            metadata.table_channel
        )
        msg = await main_table.fetch_message(record.__disco_id__)
        current = _Record.model_validate_json(msg.content).decode_content(
            record
        )

        for field, value in current.model_dump().items():
            channel = self._find_channel(
                metadata.index_channels[f"{current.__disco_name__}_{field}"]
            )

            index = self._to_index(self._hash(value))
            index_message = await self._lookup_message(channel, index)
            index_record = _IndexableRecord.from_message(index_message.content)

            if not index_record:
                raise DatabaseCorruptionError("got null record somehow")

            if len(index_record.record_ids) == 1:
                logger.info("We can nullify this entry.")
                await index_message.edit(content="null")
                self.metadata.current_records -= 1
            else:
                logger.info(
                    "There are other entries with this value, only remove this ID."  # noqa
                )
                index_record.record_ids.remove(msg.id)
                await index_message.edit(
                    content=index_record.model_dump_json(),
                )

        record.__disco_id__ = -1
        await msg.delete()



from __future__ import annotations

from datetime import datetime as dt
from enum import Enum, auto
from math import ceil

import discord
from loguru import logger

"""
How to Use:
1. Use EmbedfromContent for your database outputs with .create() and assign to a variable.
    e.g. EmbedfromContent(title="something", content={"foo":"bar"}, headers=["foo"], style="TABLE").create()
2. Use Arrow buttons class with EmbedfromContent output as content argument and assign to variable.
3. Input the ArrowButton class as the view, and the embeds as the content in interaction.send_message.
"""

__all__ = ["ArrowButtons", "EmbedFromContent", "EmbedStyle"]


class ArrowButtons(discord.ui.View):
    def __init__(self, content: list[discord.Embed]) -> None:
        super().__init__(timeout=None)
        self.value = None
        self.content = content
        self.position = 0
        self.pages = len(self.content)
        logger.debug(f"pages in button {self.pages}")
        self.on_ready()

    @discord.ui.button(
        label="◀", style=discord.ButtonStyle.primary, custom_id="l_button"
    )
    async def back(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        """Controls the left button on the qotd list embed"""
        # move back a position in the embed list
        self.position -= 1

        # check if we're on the first page, then disable the button to go left if we are (cant go anymore left)
        if self.position == 0:
            button.disabled = True

        # set the right button to a variable
        right_button = [x for x in self.children if x.custom_id == "r_button"][
            0
        ]

        # check if we're not on the last page, if yes then enable right button
        if not self.position == self.pages - 1:
            right_button.disabled = False

        # update discord message
        await interaction.response.edit_message(
            embed=self.content[self.position], view=self
        )

    @discord.ui.button(
        label="▶", style=discord.ButtonStyle.primary, custom_id="r_button"
    )
    async def forward(
        self, interaction: discord.Interaction, button: discord.ui.Button
    ) -> None:
        """Controls the right button on the qotd list embed"""
        # move forward a position in the embed list
        self.position += 1

        # set a variable for left button
        left_button = [x for x in self.children if x.custom_id == "l_button"][0]
        # check if we're not on the first page, if yes then enable left button
        if not self.position == 0:
            left_button.disabled = False

        # check if we're on the last page, if yes then disable right button
        if self.position == self.pages - 1:
            button.disabled = True

        # update discord message
        await interaction.response.edit_message(
            embed=self.content[self.position], view=self
        )

    def on_ready(self) -> None:
        """Checks the number of pages to decide which buttons to have enabled/disabled"""
        left_button = [x for x in self.children if x.custom_id == "l_button"][0]
        right_button = [x for x in self.children if x.custom_id == "r_button"][
            0
        ]

        # if we only have one page, disable both buttons
        if self.pages == 1:
            left_button.disabled = True
            right_button.disabled = True
        # if we have more than one page, only disable the left button for the first page
        else:
            left_button.disabled = True


class EmbedStyle(str, Enum):
    COLUMN = auto()
    TABLE = auto()
    SCHEMA = auto()
    DEFAULT = auto()


# TODO add support for character limits: https://anidiots.guide/.gitbook/assets/first-bot-embed-example.png
class EmbedFromContent:
    """Creates a list of embeds suited for pagination from inserted content."""

    def __init__(
        self,
        title: str,
        content: list[str] | dict | list[dict],
        style: "EmbedStyle",
        headers: list[str] | None = None,
    ) -> None:
        """
        Sets the base parameters for the embeds.

        :param title: Title of the embed.
        :param headers: Columns of the table, will be used for field names. Required if seeking table display.
        :param content: Content of the table.
        """
        self.author = "Discobase"
        self.color = discord.Colour.blurple()
        self.title = title if len(title) < 256 else f"{title[0:253]}..."
        self.headers = headers
        self.content = content
        self.page_number = 0
        self.page_total = 0
        self.url = "https://github.com/ZeroIntensity/discobase"
        self.icon_url = "https://i.imgur.com/2QH3tEQ.png"

        self.style = style

    def create(self) -> list[discord.Embed] | discord.Embed:
        if self.style == "column":
            return self._column_display()
        elif self.style == "table":
            return self._table_display()
        elif self.style == "schema":
            return self._schema_display()
        elif self.style == "default":
            return self._default_display()
        else:
            raise ValueError("Invalid style input.")

    def _column_display(self) -> list[discord.Embed]:
        """
        Creates list of discord embeds for the column content, 15 rows per embed.
        """
        entries_per_page = 15
        embeds: list[discord.Embed] = []

        column_data: list[str] = self.content
        self.page_total = ceil(len(column_data) / 15)
        logger.debug(f"{self.page_total}, round: {len(column_data) / 15}")

        # Create each embed with the data
        for i in range(0, len(column_data), entries_per_page):
            self.page_number += 1
            embed_content = "\n".join(column_data[i : i + entries_per_page])
            discord_embed = discord.Embed(
                color=self.color,
                title=self.title,
                type="rich",
                description=embed_content,
                timestamp=dt.now(),
            )
            discord_embed.set_author(
                name=self.author, url=self.url, icon_url=self.icon_url
            )
            discord_embed.set_footer(
                text=f"Page: {self.page_number}/{self.page_total}"
            )

            embeds.append(discord_embed)

        return embeds

    def _table_display(self) -> list[discord.Embed]:
        """
        Creates a list of discord embeds that display the data in a table, 10 entries per page.
        """
        entries_per_page = 10
        embeds: list[discord.Embed] = []

        column_names: list = self.headers
        table_data: dict = self.content
        self.page_total = ceil(
            len(self.content[self.headers[0]]) / entries_per_page
        )

        # get the len of the first column's data
        for i in range(0, len(table_data[column_names[0]]), entries_per_page):
            self.page_number += 1
            discord_embed = discord.Embed(
                color=self.color,
                title=self.title,
                type="rich",
                timestamp=dt.now(),
            )
            discord_embed.set_author(
                name=self.author, url=self.url, icon_url=self.icon_url
            )
            discord_embed.set_footer(
                text=f"Page: {self.page_number}/{self.page_total}"
            )
            # create fields for each column with 10 data entries
            for k, v in table_data.items():
                field_title = k.title()
                field_content = "\n".join(
                    [
                        f"**{i + 1}.** {value}"
                        for i, value in enumerate(v[i : i + entries_per_page])
                    ]
                )
                discord_embed.add_field(
                    name=field_title, value=field_content, inline=True
                )
            embeds.append(discord_embed)

        return embeds

    def _schema_display(self) -> discord.Embed:
        """
        Creates an embed that has the schema information. Column names as field titles, and type as field values.
        """
        embed = discord.Embed(
            title=self.title, color=self.color, type="rich", timestamp=dt.now()
        )
        embed.set_author(name=self.author, url=self.url, icon_url=self.icon_url)

        for content in self.content:
            embed.add_field(
                name=content["title"], value=content["type"], inline=True
            )

        return embed

    def _default_display(self) -> discord.Embed:
        """
        Creates an embed with a default visual style.
        """
        embed = discord.Embed(
            title=self.title, color=self.color, type="rich", timestamp=dt.now()
        )
        embed.set_author(name=self.author, url=self.url, icon_url=self.icon_url)

        return embed



"""
Discobase - Relation database library using nothing but Discord.

Python Discord Codejam 2024
"""

from .__about__ import __license__, __version__
from .database import *
from .exceptions import *
from .table import *



from __future__ import annotations

import asyncio
from typing import (TYPE_CHECKING, Any, ClassVar, Literal, Optional, Set,
                    overload)

import discord
from pydantic import BaseModel, ConfigDict
from typing_extensions import Self, Unpack

from ._util import free_fly

if TYPE_CHECKING:
    from .database import Database

from ._cursor import TableCursor
from .exceptions import (DatabaseLookupError, DatabaseStorageError,
                         DatabaseTableError, NotConnectedError)

__all__ = ("Table",)


# Note that we can't use 3.10+ type[] syntax
# here, since Pydantic can't handle it
class Table(BaseModel):
    __disco_database__: ClassVar[Optional[Database]]
    """Attached `Database` object. Set by the `table()` decorator."""
    __disco_cursor__: ClassVar[Optional[TableCursor]]
    """Internal table cursor, set at initialization time."""
    __disco_keys__: ClassVar[Set[str]]
    """All keys of the table, this may not change once set by `table()`."""
    __disco_name__: ClassVar[str]
    """Internal name of the table. Set by the `table()` decorator."""
    __disco_id__: int = -1
    """Message ID of the record. This is only present if it was saved."""

    def __init__(self, /, **data: Any) -> None:
        super().__init__(**data)
        self.__disco_id__ = -1

    def __init_subclass__(cls, **kwargs: Unpack[ConfigDict]) -> None:
        super().__init_subclass__(**kwargs)
        cls.__disco_database__ = None
        cls.__disco_cursor__ = None
        cls.__disco_keys__ = set()
        cls.__disco_name__ = "_notset"

    @classmethod
    def _ensure_db(cls) -> None:
        if not cls.__disco_database__:
            raise DatabaseTableError(
                f"{cls.__name__} has no attached database, did you forget to call @db.table?"  # noqa
            )

        if not cls.__disco_database__.open:
            raise NotConnectedError(
                "database is not connected! did you forget to open it?"
            )

        if not cls.__disco_cursor__:
            raise DatabaseTableError(
                f"{cls.__name__} is not ready, you might want to add a call to build_tables()",  # noqa
            )

    def save(self) -> asyncio.Task[discord.Message]:
        """
        Save the entry to the database as a new record.

        Example:
            ```py
            import discobase

            db = discobase.Database("My database")

            @db.table
            class User(discobase.Table):
                name: str
                password: str

            # Using top-level await for this example
            await User(name="Peter", password="foobar").save()
            ```
        """
        self._ensure_db()
        assert self.__disco_cursor__

        if self.__disco_id__ != -1:
            raise DatabaseStorageError(
                "this entry has already been written, did you mean to call update()?",  # noqa
            )
        task = free_fly(self.__disco_cursor__.add_record(self))

        def _cb(fut: asyncio.Task[discord.Message]) -> None:
            msg = fut.result()
            self.__disco_id__ = msg.id

        task.add_done_callback(_cb)
        return task

    def _ensure_written(self) -> None:
        if self.__disco_id__ == -1:
            raise DatabaseStorageError(
                "this entry has not been written, did you mean to call save()?",  # noqa
            )

    def update(self) -> asyncio.Task[discord.Message]:
        """
        Update the entry in-place.

        Example:
            ```py
            import discobase

            db = discobase.Database("My database")

            @db.table
            class User(discobase.Table):
                name: str
                password: str

            # Using top-level await for this example
            user = await User.find_unique(name="Peter", password="foobar")
            user.password = str(hash(password))
            await user.update()
            ```
        """

        self._ensure_db()
        self._ensure_written()
        assert self.__disco_cursor__
        if self.__disco_id__ == -1:
            raise DatabaseStorageError(
                "this entry has not been written, did you mean to call save()?",  # noqa
            )
        return free_fly(self.__disco_cursor__.update_record(self))

    def commit(self) -> asyncio.Task[discord.Message]:
        """
        Save the current entry, or update it if it already exists in the
        database.
        """
        if self.__disco_id__ == -1:
            return self.save()
        else:
            return self.update()

    def delete(self) -> asyncio.Task[None]:
        """
        Delete the current entry from the database.
        """

        self._ensure_written()
        assert self.__disco_cursor__
        return free_fly(self.__disco_cursor__.delete_record(self))

    @classmethod
    async def find(cls, **kwargs: Any) -> list[Self]:
        """
        Find a list of entries in the database.

        Args:
            **kwargs: Values to search for. These should be keys in the schema.

        Returns:
            list[Table]: The list of objects that match the values in `kwargs`.

        Example:
            ```py
            import discobase
            db = discobase.Database("My database")
            @db.table
            class User(discobase.Table):
                name: str
                password: str
            # Using top-level await for this example
            await User.find(password="foobar").save()
            ```
        """
        cls._ensure_db()
        assert cls.__disco_cursor__
        return await cls.__disco_cursor__.find_records(
            cls,
            kwargs,
        )

    @classmethod
    @overload
    async def find_unique(
        cls,
        *,
        strict: Literal[True] = True,
        **kwargs: Any,
    ) -> Self: ...

    @classmethod
    @overload
    async def find_unique(
        cls,
        *,
        strict: Literal[False] = False,
        **kwargs: Any,
    ) -> Self | None: ...

    @classmethod
    async def find_unique(
        cls,
        *,
        strict: bool = True,
        **kwargs: Any,
    ) -> Self | None:
        """
        Find a unique entry in the database.

        Args:
            **kwargs: Values to search for. These should be keys in the schema.

        Returns:
            Table: Returns a single object that matches the values in`kwargs`.
            None: Nothing was found, and `strict` is `False`.
        """

        if not kwargs:
            raise ValueError("a query must be passed to find_unique")

        values: list[Self] = await cls.find(**kwargs)

        if not len(values):
            if strict:
                raise DatabaseLookupError(
                    f"no entry found with query {kwargs}",
                )

            return None

        if strict and (1 < len(values)):
            raise DatabaseLookupError(
                "more than one entry was found with find_unique"
            )

        return values[0]



import json

import discord
from discord import app_commands
from discord.ext import commands
from loguru import logger
from pydantic import ValidationError

from ..ui import embed as em


class Utility(commands.Cog):
    """
    All the slash commands for querying information from the database.
    """

    def __init__(self, bot: commands.Bot) -> None:
        self.bot = bot
        self.db = self.bot.db

    @app_commands.command(description="Insert new data into a table.")
    @app_commands.describe(
        table="Choose the table you want to insert the data into.",
        data="The data that is to be inserted.",
    )
    async def insert(
        self,
        interaction: discord.Interaction,
        table: discord.TextChannel,
        data: str,
    ) -> None:
        logger.info("Called 'insert' command.")
        await interaction.response.send_message(
            content=f"Looking for `{table.name}`..."
        )

        table_name = table.name.replace("-", " ").lower()

        try:
            table_obj = self.db.tables[table_name]  # Table object
        except IndexError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The table `{table_name}` does not exist."
            )
            return

        try:
            await interaction.edit_original_response(
                content=f"Table `{table_name}` found! Adding data to table..."
            )
            data_dict: dict = json.loads(data)
        except TypeError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The data you entered was not in json format.\nEntered data: {data}"
            )
            return

        try:
            logger.info("Adding new data to table")
            new_entry = table_obj(**data_dict)
        except ValidationError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"You are missing one of the following columns in your data: `{table_obj.__disco_keys__}`."
            )
            return

        await new_entry.save()

        await interaction.edit_original_response(
            content=f"I have inserted `{data}` into `{table_name}` table."
        )

    @app_commands.command(
        description="Finds a record with the specific column and value in the table."
    )
    @app_commands.describe(
        table="The name of the table the column is in.",
        column="The name of the column.",
        value="The value to search for.",
    )
    async def find(
        self,
        interaction: discord.Interaction,
        table: discord.TextChannel,
        column: str,
        value: str,
    ) -> None:
        table_info: list | None = None
        results: list | None = None
        column: str = column.lower()
        results_found: int | None = None
        results_str: str = ""

        logger.debug("Find slash cmd initialised.")
        await interaction.response.send_message(
            content=f"Searching for `{value}`..."
        )

        if table.name in self.bot.db.tables:
            table_info = self.bot.db.tables[table.name]

            if column in table_info.__disco_keys__:
                results = await table_info.find(**{column: value})
                results_found = len(results)
                if results_found > 0:
                    for count, value in enumerate(results, start=1):
                        results_str += f"**{count}**. {str(value)}\n"

                    embed = em.EmbedFromContent(
                        title=f"Search Result - {results_found} Record(s) Found",
                        content=[],
                        headers=None,
                        style=em.EmbedStyle.DEFAULT,
                    ).create()

                    embed.description = results_str

                    await interaction.edit_original_response(
                        content="", embed=embed
                    )
                else:
                    await interaction.edit_original_response(
                        content="The record could not be found."
                    )
        else:
            await interaction.edit_original_response(
                content="Either the table doesn't exist or the column doesn't exist."
            )

    @app_commands.command(description="Modifies a record with a new value.")
    @app_commands.describe(
        table="Choose the table you want to perform an update on.",
        column="Choose the column you want to update.",
        current_value="The current value saved in the column.",
        new_value="Your new information.",
    )
    async def update(
        self,
        interaction: discord.Interaction,
        table: discord.TextChannel,
        column: str,
        current_value: str,
        new_value: str,
    ) -> None:
        logger.debug("Update slash cmd initialized.")
        table_info: list | None = None
        if table.name in self.db.tables:
            await interaction.response.send_message(
                content=f"Table `{table.name}` found! Searching for record..."
            )
            table_info = self.db.tables[table.name]

            try:
                if column in table_info.__disco_keys__:
                    column_name = [
                        col
                        for col in table_info.__disco_keys__
                        if col.lower() == column.lower()
                    ][0]
                    found_table = (
                        await table_info.find(**{column_name: current_value})
                    )[0]
                    setattr(found_table, column_name, new_value)
                    found_table.update()
                    await interaction.edit_original_response(
                        content=f"Successfully updated the value of **{column}** in **{table.name}**."
                    )
                else:
                    await interaction.edit_original_response(
                        content="The column does not exist."
                    )
            except ValidationError:
                await interaction.edit_original_response(
                    content=f"`{new_value}` could not be converted to the field's data type, use `/schema` to "
                    f"check the data type of the column before trying again."
                )
        else:
            await interaction.edit_original_response(
                content="There is no table with that name, try creating a table."
            )

    @app_commands.command(description="Deletes a record from a table.")
    @app_commands.describe(
        table="The table from which you want to delete",
        record="The record you want to delete - formatted as a json.",
    )
    async def delete(
        self,
        interaction: discord.Interaction,
        table: discord.TextChannel,
        record: str,
    ) -> None:
        logger.debug("Delete cmd initialized.")
        await interaction.response.send_message(
            content=f"Searching for table `{table.name}`..."
        )

        try:
            table_obj = self.db.tables[table.name]
            await interaction.edit_original_response(
                content=f"Table `{table_obj.__disco_name__}` found! Searching for record..."
            )
        except IndexError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The table `{table.name}` does not exist."
            )
            return

        try:
            record_dict = json.loads(record)
            table_record = await table_obj.find(**record_dict)
            table_record = table_record[0]
            if table_record is None:
                raise ValueError
            await interaction.edit_original_response(
                content=f"Record `{record}` found! Deleting..."
            )
        except ValueError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"No record found for `{record}`."
            )
            return
        except TypeError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The record you entered was not in json format.\nEntered record: {record}"
            )
            return

        await table_record.delete()

        await interaction.edit_original_response(
            content=f"Record `{record}` has been deleted from `{table.name}`!"
        )

    @app_commands.command(
        description="Resets the database, deleting all channels and unloading tables."
    )
    async def reset(self, interaction: discord.Interaction) -> None:
        logger.debug("Reset cmd initialized.")
        await interaction.response.send_message(
            content=f"Resetting the database, `{self.db.name}`..."
        )
        await self.db.clean()
        await interaction.edit_original_response(
            content=f"Database `{self.db.name}` has been reset!"
        )


async def setup(bot) -> None:
    await bot.add_cog(Utility(bot))



from typing import Dict, Tuple

from pydantic import BaseModel


class Metadata(BaseModel):
    name: str
    """The table name."""
    keys: Tuple[str, ...]
    """A tuple containing the name of all keys/fields of the table."""
    table_channel: int
    """Channel ID that holds the main table content."""
    index_channels: Dict[str, int]
    """A dictionary containing index channel names with index channel IDs."""
    current_records: int
    """Number of (used) records in the table."""
    max_records: int
    """Capacity of the table (i.e. the "maximum records" that is can hold)."""
    time_table: Dict[
        int, Tuple[int, int]
    ]  # Pydantic doesn't support range objects
    """Table of UNIX timestamp -> index range."""
    message_id: int
    """ID of the metadata message."""



import discord
from discord import app_commands
from discord.ext import commands
from loguru import logger

from ..ui import embed as em


class Visualization(commands.Cog):
    """
    Slash commands to visualize the database's data.
    """

    def __init__(self, bot: commands.Bot) -> None:
        self.bot = bot
        self.db = self.bot.db

    @app_commands.command()
    async def hello(self, interaction):
        await interaction.response.send_message("Hello!")

    @app_commands.command(description="View the selected table.")
    @app_commands.describe(name="The name of the table.")
    async def table(
        self, interaction: discord.Interaction, name: discord.TextChannel
    ) -> None:
        logger.debug("Table slash cmd initialized.")
        await interaction.response.send_message(
            content=f"Searching for table `{name}`..."
        )
        table_name = name.name.replace("-", " ").lower()

        try:
            table = self.db.tables[table_name]
            await interaction.edit_original_response(
                content=f"Table `{table_name}` found! Gathering data..."
            )
        except IndexError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The table `{name.name}` does not exist."
            )
            return

        table_columns = [
            col for col in table.__disco_keys__
        ]  # convert set to list to enable subscripting

        data: dict[str:list] = {}
        for col in table_columns:
            data[col] = []

        table_values = await table.find()
        logger.info(table_values)

        await interaction.edit_original_response(
            content="Still gathering data..."
        )

        for game in table_values:
            for col in table_columns:
                data[col].append(getattr(game, col))

        embed_from_content = em.EmbedFromContent(
            title=f"Table: {table.__disco_name__.title()}",
            content=data,
            headers=table_columns,
            style=em.EmbedStyle.TABLE,
        )
        embeds = embed_from_content.create()

        view = em.ArrowButtons(content=embeds)

        await interaction.edit_original_response(
            content="", embed=embeds[0], view=view
        )

    @app_commands.command(description="View the column data.")
    @app_commands.describe(
        table="The name of the table the column is in.",
        name="The name of the column.",
    )
    async def column(
        self,
        interaction: discord.Interaction,
        table: discord.TextChannel,
        name: str,
    ) -> None:
        logger.debug("Column slash cmd initialized.")
        await interaction.response.send_message(
            f"Searching for table `{table.name}`..."
        )
        try:
            col_table = self.db.tables[table.name]
            await interaction.edit_original_response(
                content=f"Table `{col_table.__disco_name__}` found! Gathering column data..."
            )
        except IndexError as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The table `{table.name}` does not exist."
            )
            return

        try:
            column = [
                col
                for col in col_table.__disco_keys__
                if col.lower() == name.lower()
            ][0]
        except (IndexError, ValueError) as e:
            logger.error(e)
            await interaction.edit_original_response(
                content=f"The column `{name}` does not exist in the table `{col_table.__disco_name__}`."
            )
            return

        table_records = await col_table.find()
        column_values = [getattr(record, column) for record in table_records]

        embeds = em.EmbedFromContent(
            title=f"Column `{name.title()}` from Table `{col_table.__disco_name__.title()}`",
            content=column_values,
            headers=None,
            style=em.EmbedStyle.COLUMN,
        ).create()

        view = em.ArrowButtons(content=embeds)

        await interaction.edit_original_response(
            content="", embed=embeds[0], view=view
        )

    @app_commands.command(
        description="Displays the number of tables and the names of the tables."
    )
    async def tablestats(self, interaction: discord.Interaction) -> None:
        logger.debug("Tablestats slash cmd initialized.")
        await interaction.response.send_message(
            content="Getting table statistics..."
        )
        try:
            tables_names: list | None = None
            tables_names = [table for table in self.db.tables]
            logger.debug(tables_names)
            combined_tables_names = "\n".join(tables_names)

            embed_gen = em.EmbedFromContent(
                title="Tables",
                content=[],
                headers=None,
                style=em.EmbedStyle.DEFAULT,
            ).create()

            embed_gen.add_field(
                name="Number of Tables",
                value=len(self.db.tables),
            )

            embed_gen.add_field(
                name="Names of Tables",
                value=combined_tables_names,
            )

            await interaction.edit_original_response(
                content="", embed=embed_gen
            )
        except Exception as e:
            logger.exception(e)
            return

    @app_commands.command(
        description="Retrieves and displays the schema for the table.",
    )
    @app_commands.describe(
        table="Choose the table you want to retrieve the schema from.",
    )
    async def schema(
        self, interaction: discord.Interaction, table: discord.TextChannel
    ) -> None:
        logger.debug("Schema slash cmd initialized.")
        await interaction.response.send_message(
            content=f"Getting schema for {table.name}..."
        )
        table_info: list | None = None
        table_schema: dict | None = None
        schemas: list[dict] | None = None
        embed_gen: discord.Embed | None = None

        if table.name in self.db.tables:
            table_info = self.db.tables[table.name]
            table_schema = table_info.model_json_schema()
            schemas = [
                table_schema["properties"][disco_key]
                for disco_key in table_info.__disco_keys__
            ]

            embed_gen = em.EmbedFromContent(
                title=f"Table: {table.name.title()}",
                content=schemas,
                headers=None,
                style=em.EmbedStyle.SCHEMA,
            ).create()

            await interaction.edit_original_response(
                content="", embed=embed_gen
            )
        else:
            await interaction.edit_original_response(
                content="There is no table with that name, try creating a table."
            )


async def setup(bot: commands.Bot) -> None:
    await bot.add_cog(Visualization(bot))
